﻿/*  Copyright © exc-jdbi 2021

    © Finite Field  2012

    © Finite Field  2012 is Free and Opensource Software
    
    www.vb-paradise.de/index.php/User/23082-exc-jdbi/
 
    github.com/exc-jdbi
*/

using System;
using System.Linq;
using System.Collections.Generic;

/*
    Wir wissen, dass jede "Zahl" (eigentlich sind
    es Polynome) im Galoise Körper eine Multiplikative 
    Inverse besitzt, da der Generator vom Körper eine 
    Prime ist, und somit auch die nach Euler phi(prime) 
    bestehende Anzahl an teilerfremde "Zahlen" in der 
    Menge vorhanden sind.
    Bei Erweiterungskörper ist das genau auch der Fall,
    nur dass die Zahlen grösser werden können als die 
    grundlegende Primzahl. 
    Die Teilerfremdheit wird zum Teil aufgehoben. Man 
    verhilft sich einfach so, dass man ein Irreduzibles 
    Polynom wählt, dass die Teilerfremdheit wieder aufrecht 
    hält. 
*/

namespace FiniteFields
{
  public partial class FiniteField
  {
    #region GCD Polynomial
    //**************** **************** **************** **************** 
    // GCD
    //**************** **************** **************** **************** 

    /// <summary>
    /// Gibt den GGT, Grössten gemeinsam Teiler zweier Polynome zurück
    /// </summary>
    /// <param name="polynom1">Das erste Polynom</param>
    /// <param name="polynom2">Das zweite Polynom</param>
    /// <param name="modulo">Begrenzt den Grad (Degree) eines Polynoms</param>
    /// <returns>GGT (GCD) zweier Polynome</returns>
    public static int[] FFPolynomGCD(int[] polynom1, int[] polynom2, int modulo)
    => EEA(polynom1, polynom2, modulo);

    /// <summary>
    /// Gibt den GGT, Grössten gemeinsam Teiler zweier Polynome zurück
    /// </summary>
    /// <param name="polynom1">Das erste Polynom</param>
    /// <param name="polynom2">Das zweite Polynom</param>
    /// <param name="reference">Referenz Finite Field</param>
    /// <returns></returns>
    public static int[] FFPolynomGCD(int[] polynom1, int[] polynom2, FiniteField reference)
   => EEA(polynom1, polynom2, reference.Generator);

    #endregion

    #region Coprime Polynomial
    //**************** **************** **************** **************** 
    // Coprime
    //**************** **************** **************** **************** 

    /// <summary>
    /// Prüft, ob zwei Polynome teilerfremd sind.
    /// </summary>
    /// <param name="polynom1">Das erste Polynom</param>
    /// <param name="polynom2">Das zweite Polynom</param>
    /// <param name="modulo">Begrenzt den Grad (Degree) eines Polynoms</param>
    /// <returns>True wenn teilerfremd, ansonsten false.</returns>
    public static bool FFPolynomIsCoprime(int[] polynom1, int[] polynom2, int modulo)
    => EEA(polynom1, polynom2, modulo).SequenceEqual(new[] { 1 });

    /// <summary>
    /// Prüft, ob zwei Polynome teilerfremd sind.
    /// </summary>
    /// <param name="polynom1">Das erste Polynom</param>
    /// <param name="polynom2">Das zweite Polynom</param>
    /// <param name="reference">Referenz Finite Field</param>
    /// <returns>True wenn teilerfremd, ansonsten false.</returns>
    public static bool FFPolynomIsCoprime(int[] polynom1, int[] polynom2, FiniteField reference)
    => FFPolynomIsCoprime(polynom1, polynom2, reference.Generator);

    #endregion

    #region Modulo Polynomial

    //**************** **************** **************** **************** 
    // Modulo
    //**************** **************** **************** **************** 

    /// <summary>
    /// Gibt das Modulo eines Polynoms zurück.
    /// </summary>
    /// <param name="dividend">Der Dividend zweier Polynome</param>
    /// <param name="divisormodulo">Das Modulo zweier Polynome</param>
    /// <param name="referencemodulo">
    /// Das zugrunde liegende Referenzmodulo.
    /// <para></para>
    /// Begrenzt den Grad (Degree) eines Polynoms 
    /// </param>
    /// <returns></returns>
    private static int[] FFPolynomModulo(
      int[] dividend, int[] divisormodulo, int referencemodulo)
    => FFPolynomDivision(dividend, divisormodulo, referencemodulo).Item2;
    #endregion

    #region Extended Euclidean Algorithm Polynomial
    //**************** **************** **************** **************** 
    // Extended Euclidean Algorithm
    //**************** **************** **************** **************** 

    /// <summary>
    /// Berechnet mithilfe des 'Erweiterten Euklid' zwei Koeffizienten,
    /// die den GGT als ganzzahlige Linearkombination ausgedrücken lassen.
    /// </summary>
    /// <param name="polynom">Das Polynom</param>
    /// <param name="reference">Referenz Finite Field</param>
    /// <returns>
    /// Gibt das Resultat als 3-er Tubel zurück<para></para>
    /// Wert1: GGT<para></para>
    /// Wert2: Koeffizient zum Polynom<para></para>
    /// Wert3: Koeffizient zum zweiten Polynom, hier meistens das Irreduzible Polnom.
    /// </returns>
    private static (int[], int[], int[]) FFPolynomExtEEa(int[] polynom, FiniteField reference)
    {
      var one = new[] { 1 };
      var zero = new[] { 0 };
      var idp = ToPolynom(reference.Polynomial, reference);

      //Sofern das Polynom kleiner als der Generator ist, 
      //kann gleich eine nummerische Berechnung gemacht werden.
      if (reference.Exponent == 1 || polynom.Length == 1)
        return FFPolynomExtEEaNumeral(polynom, reference);

      if (polynom.SequenceEqual(zero)) return (zero, zero.ToArray(), zero.ToArray());
      if (polynom.SequenceEqual(one)) return (one, one.ToArray(), idp);

      //Das Irreduzible Polynom ist normalerweise immer das grössere
      //Damit das Resultat von a in Item2 ist, genau so rein lassen.
      return FFPolynomExtEEa(polynom, idp, reference);
    }

    /// <summary>
    /// Berechnet NUMMERISCH Mithilfe des 'Erweiterten Euklid' das 
    /// Multiplikative Inverse eines Polynoms im Galoiskörper GF(Prime)
    /// </summary>
    /// <param name="finitefield">Das Polynom im GF(Prime)</param>
    /// <returns>Gibt einen nummerischen Wert zurück.</returns>
    private static FiniteField FFPolynomExtEEaNumeral(FiniteField finitefield)
    {
      if (finitefield.Exponent != 1)
        throw new InvalidOperationException("Not for Extended Fields");
      if (finitefield == 0) return finitefield;
      if (finitefield == 1) return finitefield;

      var one = new[] { 1 };
      var zero = new[] { 0 };
      var aa = finitefield.Value;
      var bb = finitefield.Generator;
      var exteea = ExtGcdIterativ(bb, aa);
      return new FiniteField(exteea.Item3, finitefield);
    }

    /// <summary>
    /// Berechnet NUMMERISCH Mithilfe des 'Erweiterten Euklid' zwei Koeffizienten,
    /// die den GGT als "ganzzahlige" Linearkombination ausgedrücken lassen.
    /// </summary>
    /// <param name="polynom">Das Polynom im GF(Prime)</param>
    /// <param name="reference">Das Referenz Finite Field</param>
    /// <returns>
    /// Gibt das Resultat als 3-er Tubel zurück<para></para>
    /// Wert1: GGT<para></para>
    /// Wert2: Koeffizient zum Polynom<para></para>
    /// Wert3: Koeffizient zum zweiten Polynom, hier meistens das Irreduzible Polnom.
    /// </returns>
    private static (int[], int[], int[]) FFPolynomExtEEaNumeral(int[] polynom, FiniteField reference)
    {
      if (reference.Exponent != 1 && polynom.Length != 1)
        throw new InvalidOperationException("Not for Extended Fields");
      var one = new[] { 1 };
      var zero = new[] { 0 };
      if (polynom.SequenceEqual(zero)) return (zero, zero.ToArray(), zero.ToArray());
      if (polynom.SequenceEqual(one)) return (one, one.ToArray(), ToPolynom(reference.Generator, reference));

      var aa = FromPolynom(polynom, reference);
      var bb = reference.Generator;
      var flag = bb >= aa;
      if (!flag) Swap(ref bb, ref aa);
      var exteea = ExtGcdIterativ(bb, aa);

      //Nur positive Werte
      exteea.Item2 = ExtMod(exteea.Item2, reference.Generator);
      exteea.Item3 = ExtMod(exteea.Item3, reference.Generator);

      //Das Resultat (Koeffizient) für das Polynom soll immer im Item2 sein,
      //damit eine Richtung gegeben ist, wo das Resultat zu finden ist.
      if (flag)
        return (ToPolynom(exteea.Item1, reference),
                ToPolynom(exteea.Item3, reference),
                ToPolynom(exteea.Item2, reference));

      return (ToPolynom(exteea.Item1, reference),
              ToPolynom(exteea.Item2, reference),
              ToPolynom(exteea.Item3, reference));
    }

    /// <summary>
    /// Berechnet polynominal Mithilfe des 'Erweiterten Euklid' zwei Koeffizienten,
    /// die den GGT als "ganzzahlige" Linearkombination ausgedrücken lassen.
    /// </summary>
    /// <param name="a">Polynom1</param>
    /// <param name="b">Polynom2</param>
    /// <param name="reference">Referenz Finite Field.</param>
    /// <returns>
    /// Gibt das Resultat als 3-er Tubel zurück<para></para>
    /// Wert1: GGT<para></para>
    /// Wert2: Koeffizient zum ersten Polynom<para></para>
    /// Wert3: Koeffizient zum zweiten Polynom, hier meistens das Irreduzible Polnom.
    /// </returns>
    private static (int[], int[], int[]) FFPolynomExtEEa(int[] a, int[] b, FiniteField reference)
    {
      int[] one = new[] { 1 };
      int[] zero = new[] { 0 };

      if (a.SequenceEqual(zero) || b.SequenceEqual(zero))
        throw new ArgumentNullException($"a / b cannot be zero.");

      if (a.SequenceEqual(one)) return (one, one.ToArray(), b);
      if (b.SequenceEqual(one)) return (one, a, one.ToArray());

      var flag = GreaterEqual(a, b);

      int[] rx1 = flag ? a.ToArray() : b.ToArray(),
            rx2 = flag ? b.ToArray() : a.ToArray(),
            ax1 = one.ToArray(),
            bx1 = zero.ToArray(),
            ax2 = zero.ToArray(),
            bx2 = one.ToArray(), z;

      var modulo = reference.Generator;
      var idp = ToPolynom(reference.Polynomial, reference);
      var q = (zero.ToArray(), new[] { -1 });
      (int[], int[], int[]) lastnozero = default;

      while (!q.Item2.SequenceEqual(one))//erster Austritt
      {
        if (q.Item2.SequenceEqual(zero)) break;//zweiter Austritt
        q = FFPolynomDivision(rx1, rx2, modulo);
        //z = FFPolynomMultiplicationInternal(ax2, q.Item1, modulo);
        z = FFPolynomMultiplication(ax2, q.Item1, idp, modulo).Item2;
        z = ExtMod(z, reference); CleanseZeros(ref z);
        z = FFPolynomAddInverse(z, reference);
        z = ExtMod(z, reference); CleanseZeros(ref z);
        ax1 = FFPolynomAdditionInternal(ax1, z, modulo);
        ax1 = ExtMod(ax1, reference); CleanseZeros(ref ax1);

        //z = FFPolynomMultiplicationInternal(bx2, q.Item1, modulo);
        z = FFPolynomMultiplication(bx2, q.Item1, idp, modulo).Item2;
        z = ExtMod(z, reference); CleanseZeros(ref z);
        z = FFPolynomAddInverse(z, reference);
        z = ExtMod(z, reference); CleanseZeros(ref z);
        bx1 = FFPolynomAdditionInternal(bx1, z, modulo);
        bx1 = ExtMod(bx1, reference); CleanseZeros(ref bx1);

        rx1 = rx2; rx2 = q.Item2;
        z = ax1; ax1 = ax2; ax2 = z;
        z = bx1; bx1 = bx2; bx2 = z;
        if (!q.Item2.SequenceEqual(zero) || lastnozero.Item1 is null)
          lastnozero = (q.Item2.ToArray(), ax2.ToArray(), bx2.ToArray());
      }

      if (q.Item2.SequenceEqual(zero))
      {
        z = FFPolynomDivision(lastnozero.Item2, lastnozero.Item1, modulo).Item1;
        z = ExtMod(z, modulo); CleanseZeros(ref z);
        lastnozero.Item2 = z;

        z = FFPolynomDivision(lastnozero.Item3, lastnozero.Item1, modulo).Item1;
        z = ExtMod(z, modulo); CleanseZeros(ref z);
        lastnozero.Item3 = z;

        lastnozero.Item1 = one;

        if (!flag)
          Swap(ref lastnozero.Item2, ref lastnozero.Item3);

        return lastnozero;
      }

      if (flag)
        return (q.Item2, ax2, bx2);

      return (q.Item2, bx2, ax2);
    }
    #endregion

    #region Euclidean Algorithm Polynomial
    //**************** **************** **************** **************** 
    // Euclidean Algorithm
    //**************** **************** **************** **************** 

    /// <summary>
    /// Berechnet Mithilfe des 'Erweiterten Euklid' den GGT zweier Polynome
    /// </summary>
    /// <param name="a">Polynom1</param>
    /// <param name="b">Polynom2</param>
    /// <param name="reference">Referenz Finite Field</param>
    /// <returns>Gibt den GGT als Polynom zurück.</returns>
    private static int[] EEA(int[] a, int[] b, FiniteField reference)
      => EEA(a, b, reference.Generator);

    /// <summary>
    /// Berechnet Mithilfe des 'Erweiterten Euklid' den GGT zweier Polynome
    /// </summary>
    /// <param name="a">Polynom1</param>
    /// <param name="b">Polynom2</param>
    /// <param name="modulo">Begrenzt den Grad (Degree) eines Polynom auf den Modulowert.</param>
    /// <returns>Gibt den GGT als Polynom zurück.</returns>
    private static int[] EEA(int[] a, int[] b, int modulo)
    {

      int[] t1, t2;
      var flag = GreaterEqual(a, b);
      t1 = flag ? a.ToArray() : b.ToArray();
      t2 = flag ? b.ToArray() : a.ToArray();
      t1 = ExtMod(t1, modulo);
      t2 = ExtMod(t2, modulo);

      var c = 2;//muss 2 sein
      int[] divresult = default;
      while (c > 1)
      {
        divresult = FFPolynomModulo(t1, t2, modulo);
        c = divresult.Length == 0 ? 0 : FromPolynom(divresult, modulo);
        t1 = t2; t2 = divresult;
      }

      if (c == 1 || c == -1)
        return divresult;

      if (c == 0)
        return t1;

      throw new InvalidOperationException("c muss 1 oder 0 ergeben");
    }
    #endregion

    #region Additive Inverses Polynomial
    //**************** **************** **************** **************** 
    // Additive Inverses
    //**************** **************** **************** **************** 

    /// <summary>
    /// Gibt zu einem Polynom das Additiven Inverse zurück.
    /// </summary>
    /// <param name="value">Polynom</param>
    /// <param name="reference">Referenz Finite Field</param>
    /// <returns>Gibt das AInv als Polynom zurück.</returns>
    public static int[] FFPolynomAddInverse(int[] value, FiniteField reference)
    {
      var val = FromPolynom(value, reference);
      return FFPolynomAdditiveInverse(val, reference);
    }

    #endregion

    #region Multiplicative Inverses Polynomial
    //**************** **************** **************** **************** 
    // Multiplicative Inverses
    //**************** **************** **************** **************** 

    /// <summary>
    /// Gibt zu einem Polynom das Additiven Inverse zurück.
    /// </summary>
    /// <param name="polynom">Polnom</param>
    /// <param name="reference">Referenz Finite Field</param>
    /// <returns>Gibt das MInv als Polynom zurück.</returns>
    public static int[] FFPolynomMultInverse(int[] polynom, FiniteField reference)
    {
      var exteea = FFPolynomExtEEa(polynom, reference);
      //Die Berechnungn sind so gemacht, dass das Resultat
      //immer im Item2 ist.
      return exteea.Item2;
    }

    /// <summary>
    /// Gibt zu einem Polynom das Additiven Inverse zurück.
    /// </summary>
    /// <param name="value">Dezimalwert Value für das Polynom</param>
    /// <param name="reference">Referenz Finite Field</param>
    /// <returns>Gibt das MInv als Polynom zurück.</returns>
    public static int[] FFPolynomMultInverse(int value, FiniteField reference)
    => FFPolynomMultInverse(ToPolynom(value, reference), reference);
    #endregion

    #region Numerical algorithms
    //**************** **************** **************** **************** 
    // Numerische Algos
    //**************** **************** **************** **************** 

    /// <summary>
    /// Extended Euclidean Algorithm - Numerisch
    /// </summary>
    /// <param name="a">Value 1</param>
    /// <param name="b">Value 2</param>
    /// <returns>GGT, Coff-A, Coff-B</returns>
    private static int[] ExtEEA(int a, int b)
    {
      a = Math.Abs(a);
      var q = new List<int>(30);
      int aa = a, bb = b, r = b;
      while (r > 0)
      {
        q.Add(a / b); r = a % b;
        a = b; b = r;
      }
      int rr = a, ax = 1, by = 0;
      for (int i = q.Count - 2; i >= 0; i--)
      {
        r = (int)(by - q[i] * (long)ax);
        by = ax; ax = r;
      }
      long check = ax * (long)aa + by * (long)bb;
      if ((check != 1L) || (check != -1))
      {
        int t = ax; ax = by; by = t;
      }
      return new int[] { rr, ax, by };
    }

    /// <summary>
    /// Extended Euclidean Algorithm - Numerisch
    /// </summary>
    /// <param name="a">Value 1</param>
    /// <param name="b">Value 2</param>
    /// <returns>GGT, Coff-A, Coff-B</returns>
    private static (int, int, int) ExtGcdIterativ(int a, int b)
    {
      int u = 1, v = 0, s = 0, t = 1, w, q;
      while (b != 0)
      {
        q = a / b;
        w = a; a = b; b = w - q * a;
        w = u; u = s; s = w - q * u;
        w = v; v = t; t = w - q * v;
      }
      return (a, u, v);
    }

    #endregion
  }
}
