﻿/*  Copyright © exc-jdbi 2021

    © Finite Field  2012

    © Finite Field  2012 is Free and Opensource Software
    
    www.vb-paradise.de/index.php/User/23082-exc-jdbi/
 
    github.com/exc-jdbi
*/


using System;
using System.Linq;

namespace FiniteFields
{
  public partial class FiniteField
  {
    /// <summary>
    /// Addition zweier FF-Elemente
    /// </summary>
    /// <param name="ffa">FF-Elemente</param>
    /// <param name="ffb">FF-Elemente</param>
    /// <returns>FF-Elemente</returns>
    public static FiniteField operator +(FiniteField ffa, FiniteField ffb)
    {
      if (ffa.Value == 0)
        return new FiniteField(ffb.Value, ffa);

      if (ffb.Value == 0)
        return new FiniteField(ffa);

      int result = 0;
      var polynomial = ToPolynom(ffa.Polynomial, ffa);
      var apolinomvalue = ToPolynom(ffa.Value, ffa);
      var bpolinomvalue = ToPolynom(ffb.Value % ffa.Order, ffa);
      var resultpolynom = FFPolynomAddition(apolinomvalue, bpolinomvalue, polynomial, ffa.Generator);
      if (ffa.Exponent == 1)
        result = FromPolynom(resultpolynom.Item1, ffa);
      else result = FromPolynom(resultpolynom.Item2, ffa);

      return new FiniteField(ffa.Generator, ffa.Exponent, ffa.Polynomial, result);
    }

    /// <summary>
    /// Subtraktion zweier FF-Elemente
    /// </summary>
    /// <param name="ffa">FF-Elemente</param>
    /// <param name="ffb">FF-Elemente</param>
    /// <returns>FF-Elemente</returns>
    public static FiniteField operator -(FiniteField ffa, FiniteField ffb)
    {
      if (ffb.Value == 0)
        return new FiniteField(ffa);

      //Sofern Basis 2, dann ist die Addition gleich der Subtraktion
      if (ffa.Generator == 2)
        return ffa + ffb;

      int result = 0;
      var polynomial = ToPolynom(ffa.Polynomial, ffa);
      var apolinomvalue = ToPolynom(ExtMod(ffa.Value, ffa.Order), ffa);

      int[] addinvers;
      if (ffa.Exponent == 1)
        addinvers = ToPolynom(ExtMod(ffa.Order - ffb.Value, ffa.Order), ffa);
      else addinvers = FFPolynomAdditiveInverse(ExtMod(ffb.Value, ffa.Order), ffa);

      var resultpolynom = FFPolynomAddition(apolinomvalue, addinvers, polynomial, ffa.Generator);
      if (ffa.Exponent == 1)
        result = ExtMod(FromPolynom(resultpolynom.Item1, ffa), ffa.Generator);
      else result = FromPolynom(resultpolynom.Item2, ffa);
      return new FiniteField(ffa.Generator, ffa.Exponent, ffa.Polynomial, result);
    }
    
    /// <summary>
    /// Multiplikation zweier FF-Elemente
    /// </summary>
    /// <param name="ffa">FF-Elemente</param>
    /// <param name="ffb">FF-Elemente</param>
    /// <returns>FF-Elemente</returns>
    public static FiniteField operator *(FiniteField ffa, FiniteField ffb)
    {
      if (ffa.Value == 0 || ffb.Value == 0)
        return new FiniteField(0, ffa);

      int result = 0;
      var polynomial = ToPolynom(ffa.Polynomial, ffa);
      var apolinomvalue = ToPolynom(ffa.Value, ffa);
      var bpolinomvalue = ToPolynom(ExtMod(ffb.Value, ffa.Order), ffa);
      var resultpolynom = FFPolynomMultiplication(apolinomvalue, bpolinomvalue, polynomial, ffa.Generator);
      //ToSetRight(resultpolynom.Item2, ffa.Generator);
      if (ffa.Exponent == 1)
        result = FromPolynom(resultpolynom.Item1, ffa);
      else result = FromPolynom(resultpolynom.Item2, ffa);
      return new FiniteField(ffa.Generator, ffa.Exponent, ffa.Polynomial, result);
    }

    /// <summary>
    /// Division zweier FF-Elemente
    /// </summary>
    /// <param name="ffa">FF-Elemente</param>
    /// <param name="ffb">FF-Elemente</param>
    /// <returns>FF-Elemente</returns>
    public static FiniteField operator /(FiniteField ffa, FiniteField ffb)
    {
      if (ffb.Value == 0)
        throw new DivideByZeroException($"Divisor cannot be 0 - {nameof(ffb)}");

      if (ffa.Value == 0)
        return new FiniteField(ffa.Generator, ffa.Exponent, ffa.Polynomial, 0);

      if (ffb.Value == 1)
        return new FiniteField(ffa);

      int result = 0;
      var polynomial = ToPolynom(ffa.Polynomial, ffa);
      var apolinomvalue = ToPolynom(ffa.Value, ffa);
      var reciprocal = FFPolynomMultInverse(ffb.Value, ffa);
      var resultpolynom = FFPolynomMultiplication(apolinomvalue, reciprocal, polynomial, ffa.Generator);
      if (ffa.Exponent == 1)
        result = FromPolynom(resultpolynom.Item1, ffa);
      else result = FromPolynom(resultpolynom.Item2, ffa);
      return new FiniteField(ffa.Generator, ffa.Exponent, ffa.Polynomial, result);
    }

    /// <summary>
    /// Macht einen Valuevergleich
    /// </summary>
    /// <param name="ffa"></param>
    /// <param name="ffb"></param>
    /// <returns></returns>
    public static bool operator ==(FiniteField ffa, FiniteField ffb)
    => ffa.Value == (ffb.Value % ffa.Order);

    /// <summary>
    /// Vergleicht die Values auf Ungleichheit.
    /// </summary>
    /// <param name="ffa"></param>
    /// <param name="ffb"></param>
    /// <returns></returns>
    public static bool operator !=(FiniteField ffa, FiniteField ffb)
    => ffa.Value != (ffb.Value % ffa.Order);

    /// <summary>
    /// Vergleicht auf Gleichheit.
    /// </summary>
    /// <param name="ffa"></param>
    /// <param name="ffb"></param>
    /// <returns></returns>
    public static bool operator ==(FiniteField ffa, int number)
    => ffa.Value == (number % ffa.Order);

    /// <summary>
    /// Vergleicht auf Ungleichheit.
    /// </summary>
    /// <param name="ffa"></param>
    /// <param name="ffb"></param>
    /// <returns></returns>
    public static bool operator !=(FiniteField ffa, int number)
    => ffa.Value != (number % ffa.Order);

    /// <summary>
    /// Macht einen Value- und Referenzvergleich
    /// </summary>
    /// <param name="obj">Das Prüf-Objekt</param>
    /// <returns>true, wenn gleich, ansonsten false</returns>
    public override bool Equals(object obj)
    {
      if (obj as FiniteField is null) return false;
      if ((object)this == obj) return true;
      return this == obj as FiniteField;
    }

    /// <summary>
    /// BKDR Methode
    /// </summary>
    /// <returns>Int32 Value</returns>
    public override int GetHashCode()
    {
      //BKDR Methode
      var cvalue = this.Value.ToString().
                   Concat(this.Order.ToString()).
                   Concat(this.Exponent.ToString()).
                   Concat(this.Generator.ToString()).ToArray();
      int seed = 131, hash = 0;
      for (int i = 0; i < cvalue.Length; i++)
        hash = (hash * seed) + ((byte)cvalue[i]);
      return hash;
    }

    public override string ToString()
    => this.Value.ToString();
  }
}
